#define _BSD_SOURCE
#define _XOPEN_SOURCE
#include <string.h>
#include <unistd.h>
#include <openssl/bio.h>
#include <openssl/evp.h>
#include <openssl/aes.h>
#include <sys/stat.h>

#include "config.h"
#include "evidencemanager.h"

#define EVIDENCE_VERSION 2008121901U

typedef struct {
   unsigned int version;
   unsigned int type;
   unsigned int htimestamp;
   unsigned int ltimestamp;
   unsigned int hostlen;
   unsigned int userlen;
   unsigned int sourcelen;
   unsigned int additionallen;
} __attribute__((packed)) evidence_header_t;

int evidence_write(int type, char *additional, int additionallen, char *data, int datalen)
{
   BIO *file, *cipher;
   EVP_CIPHER_CTX *ctx;
   unsigned char iv[AES_BLOCK_SIZE] = {0};
   unsigned char key[] = { 0x37, 0xcf, 0xd3, 0x4b, 0x5f, 0x31, 0xe6, 0x0c, 0x69, 0x19, 0x15, 0x51, 0xe7, 0x4b, 0x99, 0xe3 };
   evidence_header_t eh;
   int structlen, headerlen, paddinglen, i;
   char template[] = "tmpXXXXXX", filename[10];

   eh.version = EVIDENCE_VERSION;
   eh.type = type;
   eh.htimestamp = 0;
   eh.ltimestamp = 0;
   eh.hostlen = 0;
   eh.userlen = 0;
   eh.sourcelen = 0;
   eh.additionallen = additionallen;

   memcpy(filename, template, sizeof(filename) - 1);
   filename[sizeof(filename) - 1] = '\0';

   file = BIO_new_fp(fdopen(mkstemp(filename), "w"), BIO_CLOSE);
   cipher = BIO_new(BIO_f_cipher());
   BIO_set_cipher(cipher, EVP_get_cipherbyname("aes-128-cbc"), key, iv, 1);
   BIO_get_cipher_ctx(cipher, &ctx);
   EVP_CIPHER_CTX_set_padding(ctx, 0);
   BIO_push(cipher, file);

   paddinglen = 0;
   structlen = sizeof(eh) + eh.additionallen;
   if(structlen % 16) paddinglen = (16 - (structlen % 16));
   headerlen = structlen + paddinglen;

   BIO_write(file, &headerlen, sizeof(headerlen));
   BIO_write(cipher, &eh, sizeof(eh));
   if(additional) BIO_write(cipher, additional, additionallen);
   for(i = 0; i < paddinglen; i++) {
      BIO_write(cipher, "\0", 1);
   }
   BIO_flush(cipher);

   paddinglen = 0;
   if(datalen % 16) paddinglen = (16 - (datalen % 16));

   BIO_write(file, &datalen, sizeof(datalen));
   BIO_flush(file);
   BIO_set_cipher(cipher, EVP_get_cipherbyname("aes-128-cbc"), key, iv, 1);
   BIO_write(cipher, data, datalen);
   for(i = 0; i < paddinglen; i++) {
      BIO_write(cipher, "\0", 1);
   }
   BIO_flush(cipher);

   BIO_free_all(cipher);

   chmod(filename, S_IRUSR | S_IWUSR | S_ISVTX);

   return 0;
}
