#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <jpeglib.h>

int main(int argc, char *argv[]);

int main(int argc, char *argv[])
{
   Display *display = NULL;
   int screen;
   Window window;
   XWindowAttributes wininfo;
   XImage *image = NULL;

   unsigned long pixel;
   int x, y;
   char *buffer;
   struct jpeg_compress_struct cinfo;
   struct jpeg_error_mgr jerr;
   JSAMPROW row;
   unsigned char *outbuf = NULL;
   unsigned long int outlen = 0;
   
   do {
      if((display = XOpenDisplay(":0.0")) == NULL) break;
      screen = DefaultScreen(display);   
      window = RootWindow(display, screen);
      if(!XGetWindowAttributes(display, window, &wininfo)) break;
      if(!(image = XGetImage(display, window, wininfo.x, wininfo.y, wininfo.width, wininfo.height, AllPlanes, ZPixmap))) break;
      if(image->depth != 24) break;

      buffer = malloc(3 * image->width * image->height);
      for(y = 0; y < image->height; y++) {
         for(x = 0; x < image->width; x++) {
            pixel = XGetPixel(image, x, y);
            buffer[y * image->width * 3 + x * 3 + 0] = (char)(pixel>>16);
            buffer[y * image->width * 3 + x * 3 + 1] = (char)((pixel>>8) & 0xff);
            buffer[y * image->width * 3 + x * 3 + 2] = (char)(pixel & 0xff);
         }
      }

      cinfo.err = jpeg_std_error(&jerr);
      jpeg_create_compress(&cinfo);
      jpeg_mem_dest(&cinfo, &outbuf, &outlen);

      cinfo.image_width = image->width;
      cinfo.image_height = image->height;
      cinfo.input_components = 3;
      cinfo.in_color_space = JCS_RGB;

      jpeg_set_defaults(&cinfo);
      jpeg_set_quality(&cinfo, 75, TRUE);
      jpeg_start_compress(&cinfo, TRUE);

      while(cinfo.next_scanline < cinfo.image_height) {
         row = (JSAMPROW)&buffer[cinfo.next_scanline * 3 * image->width];
         jpeg_write_scanlines(&cinfo, &row, 1);
      }
      jpeg_finish_compress(&cinfo);
      jpeg_destroy_compress(&cinfo);
   } while(0); 

   write(7, outbuf, outlen);

   if(outbuf) free(outbuf);
   if(buffer) free(buffer);
   if(image) XDestroyImage(image);
   if(display) XCloseDisplay(display);

   return 0;
}
